from typing import Optional

import ailment

from ..goto_manager import GotoManager
from ....analyses import AnalysesHub
from ...analysis import Analysis
from ..empty_node_remover import EmptyNodeRemover
from ..jump_target_collector import JumpTargetCollector
from ..redundant_label_remover import RedundantLabelRemover
from .goto import GotoSimplifier
from .if_ import IfSimplifier
from .cascading_ifs import CascadingIfsRemover
from .ifelse import IfElseFlattener
from .loop import LoopSimplifier
from .expr_folding import ExpressionCounter, ExpressionFolder, StoreStatementFinder, ExpressionLocation
from .cascading_cond_transformer import CascadingConditionTransformer
from .switch_expr_simplifier import SwitchExpressionSimplifier
from .switch_cluster_simplifier import SwitchClusterFinder, simplify_switch_clusters, simplify_lowered_switches


class RegionSimplifier(Analysis):
    """
    Simplifies a given region.
    """

    def __init__(self, func, region, variable_kb=None, simplify_switches: bool = True, simplify_ifelse: bool = True):
        self.func = func
        self.region = region
        self.variable_kb = variable_kb
        self._simplify_switches = simplify_switches
        self._should_simplify_ifelses = simplify_ifelse

        self.goto_manager: Optional[GotoManager] = None
        self.result = None

        self._simplify()

    def _simplify(self):
        """
        RegionSimplifier performs the following simplifications:
        - Remove redundant Gotos
        - Remove redundant If/If-else statements
        """

        r = self.region
        # Remove empty nodes
        r = self._remove_empty_nodes(r)
        # Remove unnecessary Jump statements
        r = self._simplify_gotos(r)
        # Remove unnecessary jump or conditional jump statements if they jump to the successor right afterwards
        r = self._simplify_ifs(r)
        # Remove labels that are not referenced by anything
        r = self._simplify_labels(r)
        # Remove empty nodes again
        r = self._remove_empty_nodes(r)

        if self.variable_kb is not None:
            # Fold expressions that are only used once into their use sites
            r = self._fold_oneuse_expressions(r)
            r = self._remove_empty_nodes(r)

        if self._simplify_switches:
            # Simplify switch expressions
            r = self._simplify_switch_expressions(r)
            # Simplify switch clusters
            r = self._simplify_switch_clusters(r)
            # Again, remove labels that are not referenced by anything
            r = self._simplify_labels(r)

        # Remove empty nodes
        r = self._remove_empty_nodes(r)
        # Remove unnecessary else branches if the if branch will always return
        if self._should_simplify_ifelses:
            r = self._simplify_ifelses(r)
        #
        r = self._simplify_cascading_ifs(r)
        #
        r = self._simplify_loops(r)
        # Remove empty nodes again
        r = self._remove_empty_nodes(r)
        # Find nested if-else constructs and convert them into CascadingIfs
        r = self._transform_to_cascading_ifs(r)

        self.result = r

    #
    # Simplifiers
    #

    def _fold_oneuse_expressions(self, region):
        variable_manager = self.variable_kb.variables[self.func.addr]
        expr_counter = ExpressionCounter(region, variable_manager)

        variable_assignments = {}
        variable_uses = {}
        variable_assignment_dependencies = {}

        # pre-process and identify folding candidates
        # for variable definitions with loads, we invoke StoreStatementFinder to see if there are any Store statements
        # before the definition site and the use site.
        var_with_loads = {}
        single_use_variables = []
        for var, uses in expr_counter.uses.items():
            if len(uses) == 1 and var in expr_counter.assignments and len(expr_counter.assignments[var]) == 1:
                definition, deps, loc, has_loads = next(iter(expr_counter.assignments[var]))
                if has_loads:
                    # the definition has at least one load expression. we need to ensure there are no store statements
                    # between the definition site and the use site
                    _, use_expr_loc = next(iter(uses))
                    if isinstance(use_expr_loc, ExpressionLocation):
                        use_loc = use_expr_loc.statement_location()
                    else:
                        use_loc = use_expr_loc
                    var_with_loads[var] = (loc, use_loc)
                else:
                    single_use_variables.append(var)

        store_finder = StoreStatementFinder(region, set(var_with_loads.values()))
        for var, tpl in var_with_loads.items():
            if not store_finder.has_store(*tpl):
                single_use_variables.append(var)

        for var in single_use_variables:
            definition, deps, loc, _ = next(iter(expr_counter.assignments[var]))

            # make sure all variables that var depends on has been assigned at most once
            fail = False
            for dep_var in deps:
                if dep_var.is_function_argument:
                    continue
                if dep_var in expr_counter.assignments and len(expr_counter.assignments[dep_var]) > 1:
                    fail = True
                    break
            if fail:
                continue

            if isinstance(definition, ailment.Stmt.Call):
                # clear the existing variable since we no longer write to this variable after expression folding
                definition = definition.copy()
                if definition.ret_expr is not None:
                    definition.ret_expr = definition.ret_expr.copy()
                    definition.ret_expr.variable = None
            variable_assignments[var] = definition, loc
            variable_uses[var] = next(iter(expr_counter.uses[var]))
            variable_assignment_dependencies[var] = deps

        # any variable definition that uses an existing to-be-removed variable cannot be folded
        all_variables_to_fold = set(variable_assignments)
        for var in all_variables_to_fold:
            if all_variables_to_fold.intersection(variable_assignment_dependencies[var]):
                del variable_assignments[var]
                del variable_uses[var]

        # replace them
        ExpressionFolder(variable_assignments, variable_uses, region, variable_manager)
        return region

    @staticmethod
    def _simplify_switch_expressions(region):
        SwitchExpressionSimplifier(region)
        return region

    def _simplify_switch_clusters(self, region):
        finder = SwitchClusterFinder(region)
        simplify_switch_clusters(region, finder.var2condnodes, finder.var2switches)
        simplify_lowered_switches(
            region,
            {var: v for var, v in finder.var2condnodes.items() if var not in finder.var2switches},
            self.kb.functions,
        )
        return region

    @staticmethod
    def _remove_empty_nodes(region):
        return EmptyNodeRemover(region, claripy_ast_conditions=False).result

    @staticmethod
    def _transform_to_cascading_ifs(region):
        CascadingConditionTransformer(region)
        return region

    def _simplify_gotos(self, region):
        simplifier = GotoSimplifier(region, function=self.func, kb=self.kb)
        self.goto_manager = GotoManager(self.func, gotos=simplifier.irreducible_gotos)
        return region

    @staticmethod
    def _simplify_ifs(region):
        IfSimplifier(region)
        return region

    @staticmethod
    def _simplify_labels(region):
        jcl = JumpTargetCollector(region)
        RedundantLabelRemover(region, jcl.jump_targets)
        return region

    def _simplify_ifelses(self, region):
        IfElseFlattener(region, self.kb.functions)
        return region

    @staticmethod
    def _simplify_cascading_ifs(region):
        CascadingIfsRemover(region)
        return region

    def _simplify_loops(self, region):
        LoopSimplifier(region, self.kb.functions)
        return region


AnalysesHub.register_default("RegionSimplifier", RegionSimplifier)



import angr

from cle.backends.externs.simdata.io_file import io_file_data_for_arch


class fclose(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, fd_p):
        # Resolve file descriptor
        fd_offset = io_file_data_for_arch(self.state.arch)["fd"]
        fileno = self.state.mem[fd_p + fd_offset :].int.resolved

        # TODO: use a procedure that's not a linux syscall
        sys_close = angr.SIM_PROCEDURES["posix"]["close"]

        # Call system close and return
        retval = self.inline_call(sys_close, fileno).ret_expr

        return retval



from ..java import JavaSimProcedure


class Write(JavaSimProcedure):
    # TODO consider the fd

    __provides__ = (("java.io.PrintStream", "write(int)"),)

    def run(self, this, b, *args):  # pylint: disable=arguments-differ,unused-argument
        # we do % 256 since this is what Java implementation does.
        # formally what happens is %0x100 and then remove the 3 0-bytes
        # this is outputted as 1 byte only
        # I tested with: echo "a" | java -jar simple3.jar  > /tmp/o1.bin; xxd /tmp/o1.bin; 00000000: 62

        v = (b % 256).get_byte(3)
        self.state.posix.stdout.write(None, v)



DEFAULT_STATEMENT = -2
SWITCH_MISSING_DEFAULT_NODE_ADDR = 0xFFFF_FFFE
MAX_POINTSTO_BITS = -1330 * 8


def is_alignment_mask(n):
    return n in {0xFFFFFFFFFFFFFFE0, 0xFFFFFFFFFFFFFFF0, 0xFFFFFFF0, 0xFFFFFFFC}



from ..glibc.__libc_start_main import __libc_start_main as fucker


class __uClibc_main(fucker):
    # pylint: disable=missing-class-docstring
    NO_RET = True

    # This is called "fucker" cause otherwise the double underscores cause
    # python to name-mangle and everything gets fucked.



from . import ExplorationTechnique
import random


class DFS(ExplorationTechnique):
    """
    Depth-first search.

    Will only keep one path active at a time, any others will be stashed in the 'deferred' stash.
    When we run out of active paths to step, we take the longest one from deferred and continue.
    """

    def __init__(self, deferred_stash="deferred"):
        super().__init__()
        self._random = random.Random()
        self._random.seed(10)
        self.deferred_stash = deferred_stash

    def setup(self, simgr):
        if self.deferred_stash not in simgr.stashes:
            simgr.stashes[self.deferred_stash] = []

    def step(self, simgr, stash="active", **kwargs):
        simgr = simgr.step(stash=stash, **kwargs)
        if len(simgr.stashes[stash]) > 1:
            self._random.shuffle(simgr.stashes[stash])
            simgr.split(from_stash=stash, to_stash=self.deferred_stash, limit=1)

        if len(simgr.stashes[stash]) == 0:
            if len(simgr.stashes[self.deferred_stash]) == 0:
                return simgr
            simgr.stashes[stash].append(simgr.stashes[self.deferred_stash].pop())

        return simgr



# pylint:disable=too-many-boolean-expressions,global-statement
from typing import Dict, Optional, Tuple, TYPE_CHECKING
import logging

import pyvex
import archinfo


from .... import options, BP_BEFORE
from ....blade import Blade
from ....annocfg import AnnotatedCFG
from ....exploration_techniques import Slicecutor
from ....utils.constants import DEFAULT_STATEMENT
from .resolver import IndirectJumpResolver

if TYPE_CHECKING:
    from angr.block import Block


l = logging.getLogger(name=__name__)

PROFILING = False
HITS_CASE_0, HITS_CASE_1, MISSES = 0, 0, 0


def enable_profiling():
    global PROFILING, HITS_CASE_0, HITS_CASE_1, MISSES
    PROFILING = True
    HITS_CASE_0 = 0
    HITS_CASE_1 = 0
    MISSES = 0


def disable_profiling():
    global PROFILING
    PROFILING = False


class OverwriteTmpValueCallback:
    """
    Overwrites temporary values during resolution
    """

    def __init__(self, gp_value):
        self.gp_value = gp_value

    def overwrite_tmp_value(self, state):
        state.inspect.tmp_write_expr = state.solver.BVV(self.gp_value, state.arch.bits)


class MipsElfFastResolver(IndirectJumpResolver):
    """
    A timeless indirect jump resolver for R9-based indirect function calls in MIPS ELFs.
    """

    def __init__(self, project):
        super().__init__(project, timeless=True)

    def filter(self, cfg, addr, func_addr, block, jumpkind):
        if not isinstance(
            self.project.arch,
            (
                archinfo.ArchMIPS32,
                archinfo.ArchMIPS64,
            ),
        ):
            return False
        return True

    def resolve(  # pylint:disable=unused-argument
        self, cfg, addr, func_addr, block, jumpkind, func_graph_complete: bool = True, **kwargs
    ):
        """
        Wrapper for _resolve that slowly increments the max_depth used by Blade for finding sources
        until we can resolve the addr or we reach the default max_depth

        :param cfg: A CFG instance.
        :param int addr: IRSB address.
        :param int func_addr: The function address.
        :param pyvex.IRSB block: The IRSB.
        :param str jumpkind: The jumpkind.
        :return: If it was resolved and targets alongside it
        :rtype: tuple
        """
        for max_level in range(2, 4):
            resolved, resolved_targets = self._resolve(cfg, addr, func_addr, block, jumpkind, max_level=max_level)
            if resolved:
                return resolved, resolved_targets
        return False, []

    def _resolve(self, cfg, addr, func_addr, block, jumpkind, max_level):  # pylint:disable=unused-argument
        """
        Resolves the indirect jump in MIPS ELF binaries where all external function calls are indexed using gp.

        :param cfg: A CFG instance.
        :param int addr: IRSB address.
        :param int func_addr: The function address.
        :param pyvex.IRSB block: The IRSB.
        :param str jumpkind: The jumpkind.
        :param int max_level: maximum level for Blade to resolve when looking for sources
        :return: If it was resolved and targets alongside it
        :rtype: tuple
        """

        global HITS_CASE_0, HITS_CASE_1, MISSES

        project = self.project

        b = Blade(
            cfg.graph,
            addr,
            -1,
            cfg=cfg,
            project=project,
            ignore_sp=True,
            ignore_bp=True,
            ignored_regs=("gp",),
            cross_insn_opt=False,
            stop_at_calls=True,
            max_level=max_level,
            include_imarks=False,
        )

        func = cfg.kb.functions.function(addr=func_addr)
        gp_value = func.info.get("gp", None)

        # see if gp is used on this slice at all
        gp_used = self._is_gp_used_on_slice(project, b)
        if gp_used and gp_value is None:
            # this might a special case: gp is only used once in this function, and it can be initialized right
            # before its use site.
            # however, it should have been determined in CFGFast
            # cannot determine the value of gp. quit
            l.warning("Failed to determine value of register gp for function %#x.", func.addr)
            return False, []

        if gp_value is not None:
            target = self._try_handle_simple_case_0(gp_value, b)
            if target is not None:
                if PROFILING:
                    HITS_CASE_0 += 1
                    # print(f"hit/miss: {HITS_CASE_0 + HITS_CASE_1}/{MISSES}, {HITS_CASE_0}|{HITS_CASE_1}")
                return True, [target]
            target = self._try_handle_simple_case_1(gp_value, b)
            if target is not None:
                if PROFILING:
                    HITS_CASE_1 += 1
                    # print(f"hit/miss: {HITS_CASE_0 + HITS_CASE_1}/{MISSES}, {HITS_CASE_0}|{HITS_CASE_1}")
                return True, [target]

        if PROFILING:
            MISSES += 1
            # print(f"hit/miss: {HITS_CASE_0 + HITS_CASE_1}/{MISSES}, {HITS_CASE_0}|{HITS_CASE_1}")

        sources = [n for n in b.slice.nodes() if b.slice.in_degree(n) == 0]
        if not sources:
            return False, []

        source = sources[0]
        source_addr = source[0]
        annotated_cfg = AnnotatedCFG(project, None, detect_loops=False)
        annotated_cfg.from_digraph(b.slice)

        state = project.factory.blank_state(
            addr=source_addr,
            mode="fastpath",
            remove_options=options.refs,
            # suppress unconstrained stack reads for `gp`
            add_options={
                options.SYMBOL_FILL_UNCONSTRAINED_REGISTERS,
                options.SYMBOL_FILL_UNCONSTRAINED_MEMORY,
                options.NO_CROSS_INSN_OPT,
            },
        )
        state.regs._t9 = func_addr

        if gp_used:
            # Special handling for cases where `gp` is stored on the stack
            gp_offset = project.arch.registers["gp"][0]
            self._set_gp_load_callback(state, b, project, gp_offset, gp_value)
            state.regs._gp = gp_value

        simgr = self.project.factory.simulation_manager(state)
        simgr.use_technique(Slicecutor(annotated_cfg, force_sat=True))
        simgr.run()

        if simgr.cut:
            # pick the successor that is cut right after executing `addr`
            try:
                target_state = next(iter(cut for cut in simgr.cut if cut.history.addr == addr))
            except StopIteration:
                l.info("Indirect jump at %#x cannot be resolved by %s.", addr, repr(self))
                return False, []
            target = target_state.addr

            if self._is_target_valid(cfg, target) and target != func_addr:
                l.debug("Indirect jump at %#x is resolved to target %#x.", addr, target)
                return True, [target]

            l.info("Indirect jump at %#x is resolved to target %#x, which seems to be invalid.", addr, target)
            return False, []

        l.info("Indirect jump at %#x cannot be resolved by %s.", addr, repr(self))
        return False, []

    def _try_handle_simple_case_0(self, gp: int, blade: Blade) -> Optional[int]:
        # we only attempt to support the following case:
        #  + A | t37 = GET:I32(gp)
        #  + B | t36 = Add32(t37,0xffff8624)
        #  + C | t38 = LDbe:I32(t36)
        #  + D | PUT(t9) = t38
        #  + E | t8 = GET:I32(t9)
        #  Next: t8

        nodes_with_no_outedges = []
        for node in blade.slice.nodes():
            if blade.slice.out_degree(node) == 0:
                nodes_with_no_outedges.append(node)
        if len(nodes_with_no_outedges) != 1:
            return None

        end_node = nodes_with_no_outedges[0]
        if end_node[-1] != DEFAULT_STATEMENT:
            return None

        end_block = self.project.factory.block(end_node[0], cross_insn_opt=blade._cross_insn_opt).vex
        if not isinstance(end_block.next, pyvex.IRExpr.RdTmp):
            return None
        next_tmp = end_block.next.tmp

        # step backward

        # E
        previous_node = self._previous_node(blade, end_node)
        if previous_node is None:
            return None
        stmt = end_block.statements[previous_node[1]]
        if not isinstance(stmt, pyvex.IRStmt.WrTmp) or not isinstance(stmt.data, pyvex.IRExpr.Get):
            return None
        if stmt.tmp != next_tmp:
            return None
        if stmt.data.offset != self.project.arch.registers["t9"][0]:
            return None

        # D
        previous_node = self._previous_node(blade, previous_node)
        if previous_node is None:
            return None
        stmt = end_block.statements[previous_node[1]]
        if not isinstance(stmt, pyvex.IRStmt.Put) or not isinstance(stmt.data, pyvex.IRExpr.RdTmp):
            return None
        if stmt.offset != self.project.arch.registers["t9"][0]:
            return None
        data_tmp = stmt.data.tmp

        # C
        previous_node = self._previous_node(blade, previous_node)
        if previous_node is None:
            return None
        stmt = end_block.statements[previous_node[1]]
        if (
            not isinstance(stmt, pyvex.IRStmt.WrTmp)
            or not isinstance(stmt.data, pyvex.IRExpr.Load)
            or not isinstance(stmt.data.addr, pyvex.IRExpr.RdTmp)
        ):
            return None
        if stmt.tmp != data_tmp:
            return None
        addr_tmp = stmt.data.addr.tmp

        # B
        previous_node = self._previous_node(blade, previous_node)
        if previous_node is None:
            return None
        stmt = end_block.statements[previous_node[1]]
        if (
            not isinstance(stmt, pyvex.IRStmt.WrTmp)
            or stmt.tmp != addr_tmp
            or not isinstance(stmt.data, pyvex.IRExpr.Binop)
            or stmt.data.op != "Iop_Add32"
            or not isinstance(stmt.data.args[0], pyvex.IRExpr.RdTmp)
            or not isinstance(stmt.data.args[1], pyvex.IRExpr.Const)
        ):
            return None
        add_tmp = stmt.data.args[0].tmp
        add_const = stmt.data.args[1].con.value

        # A
        previous_node = self._previous_node(blade, previous_node)
        if previous_node is None:
            return None
        stmt = end_block.statements[previous_node[1]]
        if (
            not isinstance(stmt, pyvex.IRStmt.WrTmp)
            or stmt.tmp != add_tmp
            or not isinstance(stmt.data, pyvex.IRExpr.Get)
        ):
            return None
        if stmt.data.offset != self.project.arch.registers["gp"][0]:
            return None

        # matching complete
        addr = (gp + add_const) & 0xFFFF_FFFF
        try:
            target = self.project.loader.memory.unpack_word(addr, size=4)
            return target
        except KeyError:
            return None

    def _try_handle_simple_case_1(self, gp: int, blade: Blade) -> Optional[int]:
        # we only attempt to support the following case:
        #  + A | t22 = GET:I32(gp)
        #  + B | t21 = Add32(t22,0xffff8020)
        #  + C | t23 = LDbe:I32(t21)
        #  + D | PUT(t9) = t23
        #  + E | t27 = GET:I32(t9)
        #  + F | t26 = Add32(t27,0x00007cec)
        #  + G | PUT(t9) = t26
        #  + H | t4 = GET:I32(t9)
        #  + Next: t4

        nodes_with_no_outedges = []
        for node in blade.slice.nodes():
            if blade.slice.out_degree(node) == 0:
                nodes_with_no_outedges.append(node)
        if len(nodes_with_no_outedges) != 1:
            return None

        end_node = nodes_with_no_outedges[0]
        if end_node[-1] != DEFAULT_STATEMENT:
            return None

        end_block = self.project.factory.block(end_node[0], cross_insn_opt=blade._cross_insn_opt).vex
        if not isinstance(end_block.next, pyvex.IRExpr.RdTmp):
            return None
        next_tmp = end_block.next.tmp

        # step backward

        # H
        previous_node = self._previous_node(blade, end_node)
        if previous_node is None:
            return None
        stmt = end_block.statements[previous_node[1]]
        if not isinstance(stmt, pyvex.IRStmt.WrTmp) or not isinstance(stmt.data, pyvex.IRExpr.Get):
            return None
        if stmt.tmp != next_tmp:
            return None
        if stmt.data.offset != self.project.arch.registers["t9"][0]:
            return None

        # G
        previous_node = self._previous_node(blade, previous_node)
        if previous_node is None:
            return None
        stmt = end_block.statements[previous_node[1]]
        if not isinstance(stmt, pyvex.IRStmt.Put) or not isinstance(stmt.data, pyvex.IRExpr.RdTmp):
            return None
        if stmt.offset != self.project.arch.registers["t9"][0]:
            return None
        t9_tmp_G = stmt.data.tmp

        # F
        previous_node = self._previous_node(blade, previous_node)
        if previous_node is None:
            return None
        stmt = end_block.statements[previous_node[1]]
        if (
            not isinstance(stmt, pyvex.IRStmt.WrTmp)
            or not stmt.tmp == t9_tmp_G
            or not isinstance(stmt.data, pyvex.IRExpr.Binop)
            or stmt.data.op != "Iop_Add32"
            or not isinstance(stmt.data.args[0], pyvex.IRExpr.RdTmp)
            or not isinstance(stmt.data.args[1], pyvex.IRExpr.Const)
        ):
            return None
        t9_tmp_F = stmt.data.args[0].tmp
        t9_add_const = stmt.data.args[1].con.value

        # E
        previous_node = self._previous_node(blade, previous_node)
        if previous_node is None:
            return None
        stmt = end_block.statements[previous_node[1]]
        if not isinstance(stmt, pyvex.IRStmt.WrTmp) or not isinstance(stmt.data, pyvex.IRExpr.Get):
            return None
        if stmt.tmp != t9_tmp_F:
            return None
        if stmt.data.offset != self.project.arch.registers["t9"][0]:
            return None

        # D
        previous_node = self._previous_node(blade, previous_node)
        if previous_node is None:
            return None
        stmt = end_block.statements[previous_node[1]]
        if not isinstance(stmt, pyvex.IRStmt.Put) or not isinstance(stmt.data, pyvex.IRExpr.RdTmp):
            return None
        if stmt.offset != self.project.arch.registers["t9"][0]:
            return None
        t9_tmp_D = stmt.data.tmp

        # C
        previous_node = self._previous_node(blade, previous_node)
        if previous_node is None:
            return None
        stmt = end_block.statements[previous_node[1]]
        if (
            not isinstance(stmt, pyvex.IRStmt.WrTmp)
            or not isinstance(stmt.data, pyvex.IRExpr.Load)
            or not isinstance(stmt.data.addr, pyvex.IRExpr.RdTmp)
        ):
            return None
        if stmt.tmp != t9_tmp_D:
            return None
        addr_tmp = stmt.data.addr.tmp

        # B
        previous_node = self._previous_node(blade, previous_node)
        if previous_node is None:
            return None
        stmt = end_block.statements[previous_node[1]]
        if (
            not isinstance(stmt, pyvex.IRStmt.WrTmp)
            or stmt.tmp != addr_tmp
            or not isinstance(stmt.data, pyvex.IRExpr.Binop)
            or stmt.data.op != "Iop_Add32"
            or not isinstance(stmt.data.args[0], pyvex.IRExpr.RdTmp)
            or not isinstance(stmt.data.args[1], pyvex.IRExpr.Const)
        ):
            return None
        add_tmp = stmt.data.args[0].tmp
        add_const = stmt.data.args[1].con.value

        # A
        previous_node = self._previous_node(blade, previous_node)
        if previous_node is None:
            return None
        stmt = end_block.statements[previous_node[1]]
        if (
            not isinstance(stmt, pyvex.IRStmt.WrTmp)
            or stmt.tmp != add_tmp
            or not isinstance(stmt.data, pyvex.IRExpr.Get)
        ):
            return None
        if stmt.data.offset != self.project.arch.registers["gp"][0]:
            return None

        # matching complete
        addr = (gp + add_const) & 0xFFFF_FFFF
        try:
            target_0 = self.project.loader.memory.unpack_word(addr, size=4)
            target = (target_0 + t9_add_const) & 0xFFFF_FFFF
            return target
        except KeyError:
            return None

    @staticmethod
    def _previous_node(blade: Blade, curr_node: Tuple[int, int]) -> Optional[Tuple[int, int]]:
        if blade.slice.in_degree(curr_node) != 1:
            return None
        nn = next(iter(blade.slice.predecessors(curr_node)))
        if nn[0] != curr_node[0]:
            return None
        return nn

    @staticmethod
    def _set_gp_load_callback(state, blade, project, gp_offset, gp_value):
        tmps = {}
        for block_addr_in_slice in {slice_node[0] for slice_node in blade.slice.nodes()}:
            for stmt in project.factory.block(block_addr_in_slice, cross_insn_opt=False).vex.statements:
                if isinstance(stmt, pyvex.IRStmt.WrTmp) and isinstance(stmt.data, pyvex.IRExpr.Load):
                    # Load from memory to a tmp - assuming it's loading from the stack
                    tmps[stmt.tmp] = "stack"
                elif isinstance(stmt, pyvex.IRStmt.Put) and stmt.offset == gp_offset:
                    if isinstance(stmt.data, pyvex.IRExpr.RdTmp):
                        tmp_offset = stmt.data.tmp  # pylint:disable=cell-var-from-loop
                        if tmps.get(tmp_offset, None) == "stack":
                            # found the load from stack
                            # we must make sure value of that temporary variable equals to the correct gp value
                            state.inspect.make_breakpoint(
                                "tmp_write",
                                when=BP_BEFORE,
                                condition=(
                                    lambda s, bbl_addr_=block_addr_in_slice, tmp_offset_=tmp_offset: s.scratch.bbl_addr
                                    == bbl_addr_
                                    and s.inspect.tmp_write_num == tmp_offset_
                                ),
                                action=OverwriteTmpValueCallback(gp_value).overwrite_tmp_value,
                            )
                            break

    @staticmethod
    def _is_gp_used_on_slice(project, b: Blade) -> bool:
        gp_offset = project.arch.registers["gp"][0]
        blocks_on_slice: Dict[int, "Block"] = {}
        for block_addr, block_stmt_idx in b.slice.nodes():
            if block_addr not in blocks_on_slice:
                blocks_on_slice[block_addr] = project.factory.block(block_addr, cross_insn_opt=False)
            block = blocks_on_slice[block_addr]
            if block_stmt_idx == DEFAULT_STATEMENT:
                if isinstance(block.vex.next, pyvex.IRExpr.Get) and block.vex.next.offset == gp_offset:
                    gp_used = True
                    break
            else:
                stmt = block.vex.statements[block_stmt_idx]
                if (
                    isinstance(stmt, pyvex.IRStmt.WrTmp)
                    and isinstance(stmt.data, pyvex.IRExpr.Get)
                    and stmt.data.offset == gp_offset
                ):
                    gp_used = True
                    break
        else:
            gp_used = False

        return gp_used



import logging

from . import JNISimProcedure

l = logging.getLogger("angr.procedures.java_jni.global_and_local_references")

# pylint: disable=arguments-differ,unused-argument

#
# NewGlobalRef / NewWeakGlobalRef
#


class NewGlobalRef(JNISimProcedure):
    return_ty = "reference"

    def run(self, ptr_env, obj_):
        obj = self.state.jni_references.lookup(obj_)
        return self.state.jni_references.create_new_reference(obj, global_ref=True)


#
# DeleteGlobalRef / DeleteWeakGlobalRef
#


class DeleteGlobalRef(JNISimProcedure):
    return_ty = "void"

    def run(self, ptr_env, obj_):
        self.state.jni_references.delete_reference(obj_, global_ref=True)


#
# NewLocalRef
#


class NewLocalRef(JNISimProcedure):
    return_ty = "reference"

    def run(self, ptr_env, obj_):
        obj = self.state.jni_references.lookup(obj_)
        return self.state.jni_references.create_new_reference(obj)


#
# DeleteLocalRef
#


class DeleteLocalRef(JNISimProcedure):
    return_ty = "void"

    def run(self, ptr_env, obj_):
        self.state.jni_references.delete_reference(obj_)



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("rtm.dll")
prototypes = \
    {
        #
        'MgmRegisterMProtocol': SimTypeFunction([SimTypePointer(SimStruct({"dwCallbackFlags": SimTypeInt(signed=False, label="UInt32"), "pfnRpfCallback": SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwSourceAddr", "dwSourceMask", "dwGroupAddr", "dwGroupMask", "pdwInIfIndex", "pdwInIfNextHopAddr", "pdwUpStreamNbr", "dwHdrSize", "pbPacketHdr", "pbRoute"]), offset=0), "pfnCreationAlertCallback": SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"dwIfIndex": SimTypeInt(signed=False, label="UInt32"), "dwIfNextHopAddr": SimTypeInt(signed=False, label="UInt32"), "bIGMP": SimTypeInt(signed=True, label="Int32"), "bIsEnabled": SimTypeInt(signed=True, label="Int32")}, name="MGM_IF_ENTRY", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwSourceAddr", "dwSourceMask", "dwGroupAddr", "dwGroupMask", "dwInIfIndex", "dwInIfNextHopAddr", "dwIfCount", "pmieOutIfList"]), offset=0), "pfnPruneAlertCallback": SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwSourceAddr", "dwSourceMask", "dwGroupAddr", "dwGroupMask", "dwIfIndex", "dwIfNextHopAddr", "bMemberDelete", "pdwTimeout"]), offset=0), "pfnJoinAlertCallback": SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwSourceAddr", "dwSourceMask", "dwGroupAddr", "dwGroupMask", "bMemberUpdate"]), offset=0), "pfnWrongIfCallback": SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwSourceAddr", "dwGroupAddr", "dwIfIndex", "dwIfNextHopAddr", "dwHdrSize", "pbPacketHdr"]), offset=0), "pfnLocalJoinCallback": SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwSourceAddr", "dwSourceMask", "dwGroupAddr", "dwGroupMask", "dwIfIndex", "dwIfNextHopAddr"]), offset=0), "pfnLocalLeaveCallback": SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwSourceAddr", "dwSourceMask", "dwGroupAddr", "dwGroupMask", "dwIfIndex", "dwIfNextHopAddr"]), offset=0), "pfnDisableIgmpCallback": SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwIfIndex", "dwIfNextHopAddr"]), offset=0), "pfnEnableIgmpCallback": SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwIfIndex", "dwIfNextHopAddr"]), offset=0)}, name="ROUTING_PROTOCOL_CONFIG", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["prpiInfo", "dwProtocolId", "dwComponentId", "phProtocol"]),
        #
        'MgmDeRegisterMProtocol': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hProtocol"]),
        #
        'MgmTakeInterfaceOwnership': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hProtocol", "dwIfIndex", "dwIfNextHopAddr"]),
        #
        'MgmReleaseInterfaceOwnership': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hProtocol", "dwIfIndex", "dwIfNextHopAddr"]),
        #
        'MgmGetProtocolOnInterface': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwIfIndex", "dwIfNextHopAddr", "pdwIfProtocolId", "pdwIfComponentId"]),
        #
        'MgmAddGroupMembershipEntry': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hProtocol", "dwSourceAddr", "dwSourceMask", "dwGroupAddr", "dwGroupMask", "dwIfIndex", "dwIfNextHopIPAddr", "dwFlags"]),
        #
        'MgmDeleteGroupMembershipEntry': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hProtocol", "dwSourceAddr", "dwSourceMask", "dwGroupAddr", "dwGroupMask", "dwIfIndex", "dwIfNextHopIPAddr", "dwFlags"]),
        #
        'MgmGetMfe': SimTypeFunction([SimTypePointer(SimStruct({"dwGroup": SimTypeInt(signed=False, label="UInt32"), "dwSource": SimTypeInt(signed=False, label="UInt32"), "dwSrcMask": SimTypeInt(signed=False, label="UInt32"), "dwUpStrmNgbr": SimTypeInt(signed=False, label="UInt32"), "dwInIfIndex": SimTypeInt(signed=False, label="UInt32"), "dwInIfProtocol": SimTypeInt(signed=False, label="UInt32"), "dwRouteProtocol": SimTypeInt(signed=False, label="UInt32"), "dwRouteNetwork": SimTypeInt(signed=False, label="UInt32"), "dwRouteMask": SimTypeInt(signed=False, label="UInt32"), "ulUpTime": SimTypeInt(signed=False, label="UInt32"), "ulExpiryTime": SimTypeInt(signed=False, label="UInt32"), "ulTimeOut": SimTypeInt(signed=False, label="UInt32"), "ulNumOutIf": SimTypeInt(signed=False, label="UInt32"), "fFlags": SimTypeInt(signed=False, label="UInt32"), "dwReserved": SimTypeInt(signed=False, label="UInt32"), "rgmioOutInfo": SimTypePointer(SimStruct({"dwOutIfIndex": SimTypeInt(signed=False, label="UInt32"), "dwNextHopAddr": SimTypeInt(signed=False, label="UInt32"), "dwReserved": SimTypeInt(signed=False, label="UInt32"), "dwReserved1": SimTypeInt(signed=False, label="UInt32")}, name="MIB_IPMCAST_OIF_XP", pack=False, align=None), offset=0)}, name="MIB_IPMCAST_MFE", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pimm", "pdwBufferSize", "pbBuffer"]),
        #
        'MgmGetFirstMfe': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pdwBufferSize", "pbBuffer", "pdwNumEntries"]),
        #
        'MgmGetNextMfe': SimTypeFunction([SimTypePointer(SimStruct({"dwGroup": SimTypeInt(signed=False, label="UInt32"), "dwSource": SimTypeInt(signed=False, label="UInt32"), "dwSrcMask": SimTypeInt(signed=False, label="UInt32"), "dwUpStrmNgbr": SimTypeInt(signed=False, label="UInt32"), "dwInIfIndex": SimTypeInt(signed=False, label="UInt32"), "dwInIfProtocol": SimTypeInt(signed=False, label="UInt32"), "dwRouteProtocol": SimTypeInt(signed=False, label="UInt32"), "dwRouteNetwork": SimTypeInt(signed=False, label="UInt32"), "dwRouteMask": SimTypeInt(signed=False, label="UInt32"), "ulUpTime": SimTypeInt(signed=False, label="UInt32"), "ulExpiryTime": SimTypeInt(signed=False, label="UInt32"), "ulTimeOut": SimTypeInt(signed=False, label="UInt32"), "ulNumOutIf": SimTypeInt(signed=False, label="UInt32"), "fFlags": SimTypeInt(signed=False, label="UInt32"), "dwReserved": SimTypeInt(signed=False, label="UInt32"), "rgmioOutInfo": SimTypePointer(SimStruct({"dwOutIfIndex": SimTypeInt(signed=False, label="UInt32"), "dwNextHopAddr": SimTypeInt(signed=False, label="UInt32"), "dwReserved": SimTypeInt(signed=False, label="UInt32"), "dwReserved1": SimTypeInt(signed=False, label="UInt32")}, name="MIB_IPMCAST_OIF_XP", pack=False, align=None), offset=0)}, name="MIB_IPMCAST_MFE", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pimmStart", "pdwBufferSize", "pbBuffer", "pdwNumEntries"]),
        #
        'MgmGetMfeStats': SimTypeFunction([SimTypePointer(SimStruct({"dwGroup": SimTypeInt(signed=False, label="UInt32"), "dwSource": SimTypeInt(signed=False, label="UInt32"), "dwSrcMask": SimTypeInt(signed=False, label="UInt32"), "dwUpStrmNgbr": SimTypeInt(signed=False, label="UInt32"), "dwInIfIndex": SimTypeInt(signed=False, label="UInt32"), "dwInIfProtocol": SimTypeInt(signed=False, label="UInt32"), "dwRouteProtocol": SimTypeInt(signed=False, label="UInt32"), "dwRouteNetwork": SimTypeInt(signed=False, label="UInt32"), "dwRouteMask": SimTypeInt(signed=False, label="UInt32"), "ulUpTime": SimTypeInt(signed=False, label="UInt32"), "ulExpiryTime": SimTypeInt(signed=False, label="UInt32"), "ulTimeOut": SimTypeInt(signed=False, label="UInt32"), "ulNumOutIf": SimTypeInt(signed=False, label="UInt32"), "fFlags": SimTypeInt(signed=False, label="UInt32"), "dwReserved": SimTypeInt(signed=False, label="UInt32"), "rgmioOutInfo": SimTypePointer(SimStruct({"dwOutIfIndex": SimTypeInt(signed=False, label="UInt32"), "dwNextHopAddr": SimTypeInt(signed=False, label="UInt32"), "dwReserved": SimTypeInt(signed=False, label="UInt32"), "dwReserved1": SimTypeInt(signed=False, label="UInt32")}, name="MIB_IPMCAST_OIF_XP", pack=False, align=None), offset=0)}, name="MIB_IPMCAST_MFE", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["pimm", "pdwBufferSize", "pbBuffer", "dwFlags"]),
        #
        'MgmGetFirstMfeStats': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["pdwBufferSize", "pbBuffer", "pdwNumEntries", "dwFlags"]),
        #
        'MgmGetNextMfeStats': SimTypeFunction([SimTypePointer(SimStruct({"dwGroup": SimTypeInt(signed=False, label="UInt32"), "dwSource": SimTypeInt(signed=False, label="UInt32"), "dwSrcMask": SimTypeInt(signed=False, label="UInt32"), "dwUpStrmNgbr": SimTypeInt(signed=False, label="UInt32"), "dwInIfIndex": SimTypeInt(signed=False, label="UInt32"), "dwInIfProtocol": SimTypeInt(signed=False, label="UInt32"), "dwRouteProtocol": SimTypeInt(signed=False, label="UInt32"), "dwRouteNetwork": SimTypeInt(signed=False, label="UInt32"), "dwRouteMask": SimTypeInt(signed=False, label="UInt32"), "ulUpTime": SimTypeInt(signed=False, label="UInt32"), "ulExpiryTime": SimTypeInt(signed=False, label="UInt32"), "ulTimeOut": SimTypeInt(signed=False, label="UInt32"), "ulNumOutIf": SimTypeInt(signed=False, label="UInt32"), "fFlags": SimTypeInt(signed=False, label="UInt32"), "dwReserved": SimTypeInt(signed=False, label="UInt32"), "rgmioOutInfo": SimTypePointer(SimStruct({"dwOutIfIndex": SimTypeInt(signed=False, label="UInt32"), "dwNextHopAddr": SimTypeInt(signed=False, label="UInt32"), "dwReserved": SimTypeInt(signed=False, label="UInt32"), "dwReserved1": SimTypeInt(signed=False, label="UInt32")}, name="MIB_IPMCAST_OIF_XP", pack=False, align=None), offset=0)}, name="MIB_IPMCAST_MFE", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["pimmStart", "pdwBufferSize", "pbBuffer", "pdwNumEntries", "dwFlags"]),
        #
        'MgmGroupEnumerationStart': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="MGM_ENUM_TYPES"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hProtocol", "metEnumType", "phEnumHandle"]),
        #
        'MgmGroupEnumerationGetNext': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hEnum", "pdwBufferSize", "pbBuffer", "pdwNumEntries"]),
        #
        'MgmGroupEnumerationEnd': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hEnum"]),
        #
        'RtmConvertNetAddressToIpv6AddressAndLength': SimTypeFunction([SimTypePointer(SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"u": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "Word": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 8)}, name="<anon>", label="None")}, name="IN6_ADDR", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["pNetAddress", "pAddress", "pLength", "dwAddressSize"]),
        #
        'RtmConvertIpv6AddressAndLengthToNetAddress': SimTypeFunction([SimTypePointer(SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), offset=0), SimStruct({"u": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "Word": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 8)}, name="<anon>", label="None")}, name="IN6_ADDR", pack=False, align=None), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["pNetAddress", "Address", "dwLength", "dwAddressSize"]),
        #
        'RtmRegisterEntity': SimTypeFunction([SimTypePointer(SimStruct({"RtmInstanceId": SimTypeShort(signed=False, label="UInt16"), "AddressFamily": SimTypeShort(signed=False, label="UInt16"), "EntityId": SimStruct({"Anonymous": SimUnion({"Anonymous": SimStruct({"EntityProtocolId": SimTypeInt(signed=False, label="UInt32"), "EntityInstanceId": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "EntityId": SimTypeLongLong(signed=False, label="UInt64")}, name="<anon>", label="None")}, name="RTM_ENTITY_ID", pack=False, align=None)}, name="RTM_ENTITY_INFO", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"NumMethods": SimTypeInt(signed=False, label="UInt32"), "Methods": SimTypePointer(SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"MethodType": SimTypeInt(signed=False, label="UInt32"), "InputSize": SimTypeInt(signed=False, label="UInt32"), "InputData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="RTM_ENTITY_METHOD_INPUT", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"MethodType": SimTypeInt(signed=False, label="UInt32"), "MethodStatus": SimTypeInt(signed=False, label="UInt32"), "OutputSize": SimTypeInt(signed=False, label="UInt32"), "OutputData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="RTM_ENTITY_METHOD_OUTPUT", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["CallerHandle", "CalleeHandle", "Input", "Output"]), offset=0), offset=0)}, name="RTM_ENTITY_EXPORT_METHODS", pack=False, align=None), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="RTM_EVENT_TYPE"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "EventType", "Context1", "Context2"]), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimStruct({"MaxNextHopsInRoute": SimTypeInt(signed=False, label="UInt32"), "MaxHandlesInEnum": SimTypeInt(signed=False, label="UInt32"), "ViewsSupported": SimTypeInt(signed=False, label="UInt32"), "NumberOfViews": SimTypeInt(signed=False, label="UInt32")}, name="RTM_REGN_PROFILE", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmEntityInfo", "ExportMethods", "EventCallback", "ReserveOpaquePointer", "RtmRegProfile", "RtmRegHandle"]),
        #
        'RtmDeregisterEntity': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle"]),
        #
        'RtmGetRegisteredEntities': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimStruct({"RtmInstanceId": SimTypeShort(signed=False, label="UInt16"), "AddressFamily": SimTypeShort(signed=False, label="UInt16"), "EntityId": SimStruct({"Anonymous": SimUnion({"Anonymous": SimStruct({"EntityProtocolId": SimTypeInt(signed=False, label="UInt32"), "EntityInstanceId": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "EntityId": SimTypeLongLong(signed=False, label="UInt64")}, name="<anon>", label="None")}, name="RTM_ENTITY_ID", pack=False, align=None)}, name="RTM_ENTITY_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NumEntities", "EntityHandles", "EntityInfos"]),
        #
        'RtmReleaseEntities': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NumEntities", "EntityHandles"]),
        #
        'RtmLockDestination': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "DestHandle", "Exclusive", "LockDest"]),
        #
        'RtmGetOpaqueInformationPointer': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "DestHandle", "OpaqueInfoPointer"]),
        #
        'RtmGetEntityMethods': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"MethodType": SimTypeInt(signed=False, label="UInt32"), "InputSize": SimTypeInt(signed=False, label="UInt32"), "InputData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="RTM_ENTITY_METHOD_INPUT", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"MethodType": SimTypeInt(signed=False, label="UInt32"), "MethodStatus": SimTypeInt(signed=False, label="UInt32"), "OutputSize": SimTypeInt(signed=False, label="UInt32"), "OutputData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="RTM_ENTITY_METHOD_OUTPUT", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["CallerHandle", "CalleeHandle", "Input", "Output"]), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "EntityHandle", "NumMethods", "ExptMethods"]),
        #
        'RtmInvokeMethod': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"MethodType": SimTypeInt(signed=False, label="UInt32"), "InputSize": SimTypeInt(signed=False, label="UInt32"), "InputData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="RTM_ENTITY_METHOD_INPUT", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"MethodType": SimTypeInt(signed=False, label="UInt32"), "MethodStatus": SimTypeInt(signed=False, label="UInt32"), "OutputSize": SimTypeInt(signed=False, label="UInt32"), "OutputData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="RTM_ENTITY_METHOD_OUTPUT", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "EntityHandle", "Input", "OutputSize", "Output"]),
        #
        'RtmBlockMethods': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeChar(label="Byte"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "TargetHandle", "TargetType", "BlockingFlag"]),
        #
        'RtmGetEntityInfo': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"RtmInstanceId": SimTypeShort(signed=False, label="UInt16"), "AddressFamily": SimTypeShort(signed=False, label="UInt16"), "EntityId": SimStruct({"Anonymous": SimUnion({"Anonymous": SimStruct({"EntityProtocolId": SimTypeInt(signed=False, label="UInt32"), "EntityInstanceId": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "EntityId": SimTypeLongLong(signed=False, label="UInt64")}, name="<anon>", label="None")}, name="RTM_ENTITY_ID", pack=False, align=None)}, name="RTM_ENTITY_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "EntityHandle", "EntityInfo"]),
        #
        'RtmGetDestInfo': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"DestHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestAddress": SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), "LastChanged": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "BelongsToViews": SimTypeInt(signed=False, label="UInt32"), "NumberOfViews": SimTypeInt(signed=False, label="UInt32"), "ViewInfo": SimTypePointer(SimStruct({"ViewId": SimTypeInt(signed=True, label="Int32"), "NumRoutes": SimTypeInt(signed=False, label="UInt32"), "Route": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestFlags": SimTypeInt(signed=False, label="UInt32"), "HoldRoute": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="_Anonymous_e__Struct", pack=False, align=None), offset=0)}, name="RTM_DEST_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "DestHandle", "ProtocolId", "TargetViews", "DestInfo"]),
        #
        'RtmGetRouteInfo': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"DestHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "RouteOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Neighbour": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "State": SimTypeChar(label="Byte"), "Flags1": SimTypeChar(label="Byte"), "Flags": SimTypeShort(signed=False, label="UInt16"), "PrefInfo": SimStruct({"Metric": SimTypeInt(signed=False, label="UInt32"), "Preference": SimTypeInt(signed=False, label="UInt32")}, name="RTM_PREF_INFO", pack=False, align=None), "BelongsToViews": SimTypeInt(signed=False, label="UInt32"), "EntitySpecificInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "NextHopsList": SimStruct({"NumNextHops": SimTypeShort(signed=False, label="UInt16"), "NextHops": SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)}, name="RTM_NEXTHOP_LIST", pack=False, align=None)}, name="RTM_ROUTE_INFO", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "RouteHandle", "RouteInfo", "DestAddress"]),
        #
        'RtmGetNextHopInfo': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"NextHopAddress": SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), "NextHopOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "InterfaceIndex": SimTypeInt(signed=False, label="UInt32"), "State": SimTypeShort(signed=False, label="UInt16"), "Flags": SimTypeShort(signed=False, label="UInt16"), "EntitySpecificInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "RemoteNextHop": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="RTM_NEXTHOP_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NextHopHandle", "NextHopInfo"]),
        #
        'RtmReleaseEntityInfo': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"RtmInstanceId": SimTypeShort(signed=False, label="UInt16"), "AddressFamily": SimTypeShort(signed=False, label="UInt16"), "EntityId": SimStruct({"Anonymous": SimUnion({"Anonymous": SimStruct({"EntityProtocolId": SimTypeInt(signed=False, label="UInt32"), "EntityInstanceId": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "EntityId": SimTypeLongLong(signed=False, label="UInt64")}, name="<anon>", label="None")}, name="RTM_ENTITY_ID", pack=False, align=None)}, name="RTM_ENTITY_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "EntityInfo"]),
        #
        'RtmReleaseDestInfo': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"DestHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestAddress": SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), "LastChanged": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "BelongsToViews": SimTypeInt(signed=False, label="UInt32"), "NumberOfViews": SimTypeInt(signed=False, label="UInt32"), "ViewInfo": SimTypePointer(SimStruct({"ViewId": SimTypeInt(signed=True, label="Int32"), "NumRoutes": SimTypeInt(signed=False, label="UInt32"), "Route": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestFlags": SimTypeInt(signed=False, label="UInt32"), "HoldRoute": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="_Anonymous_e__Struct", pack=False, align=None), offset=0)}, name="RTM_DEST_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "DestInfo"]),
        #
        'RtmReleaseRouteInfo': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"DestHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "RouteOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Neighbour": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "State": SimTypeChar(label="Byte"), "Flags1": SimTypeChar(label="Byte"), "Flags": SimTypeShort(signed=False, label="UInt16"), "PrefInfo": SimStruct({"Metric": SimTypeInt(signed=False, label="UInt32"), "Preference": SimTypeInt(signed=False, label="UInt32")}, name="RTM_PREF_INFO", pack=False, align=None), "BelongsToViews": SimTypeInt(signed=False, label="UInt32"), "EntitySpecificInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "NextHopsList": SimStruct({"NumNextHops": SimTypeShort(signed=False, label="UInt16"), "NextHops": SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)}, name="RTM_NEXTHOP_LIST", pack=False, align=None)}, name="RTM_ROUTE_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "RouteInfo"]),
        #
        'RtmReleaseNextHopInfo': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"NextHopAddress": SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), "NextHopOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "InterfaceIndex": SimTypeInt(signed=False, label="UInt32"), "State": SimTypeShort(signed=False, label="UInt16"), "Flags": SimTypeShort(signed=False, label="UInt16"), "EntitySpecificInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "RemoteNextHop": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="RTM_NEXTHOP_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NextHopInfo"]),
        #
        'RtmAddRouteToDest': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"DestHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "RouteOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Neighbour": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "State": SimTypeChar(label="Byte"), "Flags1": SimTypeChar(label="Byte"), "Flags": SimTypeShort(signed=False, label="UInt16"), "PrefInfo": SimStruct({"Metric": SimTypeInt(signed=False, label="UInt32"), "Preference": SimTypeInt(signed=False, label="UInt32")}, name="RTM_PREF_INFO", pack=False, align=None), "BelongsToViews": SimTypeInt(signed=False, label="UInt32"), "EntitySpecificInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "NextHopsList": SimStruct({"NumNextHops": SimTypeShort(signed=False, label="UInt16"), "NextHops": SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)}, name="RTM_NEXTHOP_LIST", pack=False, align=None)}, name="RTM_ROUTE_INFO", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "RouteHandle", "DestAddress", "RouteInfo", "TimeToLive", "RouteListHandle", "NotifyType", "NotifyHandle", "ChangeFlags"]),
        #
        'RtmDeleteRouteToDest': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "RouteHandle", "ChangeFlags"]),
        #
        'RtmHoldDestination': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "DestHandle", "TargetViews", "HoldTime"]),
        #
        'RtmGetRoutePointer': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimStruct({"DestHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "RouteOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Neighbour": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "State": SimTypeChar(label="Byte"), "Flags1": SimTypeChar(label="Byte"), "Flags": SimTypeShort(signed=False, label="UInt16"), "PrefInfo": SimStruct({"Metric": SimTypeInt(signed=False, label="UInt32"), "Preference": SimTypeInt(signed=False, label="UInt32")}, name="RTM_PREF_INFO", pack=False, align=None), "BelongsToViews": SimTypeInt(signed=False, label="UInt32"), "EntitySpecificInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "NextHopsList": SimStruct({"NumNextHops": SimTypeShort(signed=False, label="UInt16"), "NextHops": SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)}, name="RTM_NEXTHOP_LIST", pack=False, align=None)}, name="RTM_ROUTE_INFO", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "RouteHandle", "RoutePointer"]),
        #
        'RtmLockRoute': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypePointer(SimStruct({"DestHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "RouteOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Neighbour": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "State": SimTypeChar(label="Byte"), "Flags1": SimTypeChar(label="Byte"), "Flags": SimTypeShort(signed=False, label="UInt16"), "PrefInfo": SimStruct({"Metric": SimTypeInt(signed=False, label="UInt32"), "Preference": SimTypeInt(signed=False, label="UInt32")}, name="RTM_PREF_INFO", pack=False, align=None), "BelongsToViews": SimTypeInt(signed=False, label="UInt32"), "EntitySpecificInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "NextHopsList": SimStruct({"NumNextHops": SimTypeShort(signed=False, label="UInt16"), "NextHops": SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)}, name="RTM_NEXTHOP_LIST", pack=False, align=None)}, name="RTM_ROUTE_INFO", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "RouteHandle", "Exclusive", "LockRoute", "RoutePointer"]),
        #
        'RtmUpdateAndUnlockRoute': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "RouteHandle", "TimeToLive", "RouteListHandle", "NotifyType", "NotifyHandle", "ChangeFlags"]),
        #
        'RtmGetExactMatchDestination': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"DestHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestAddress": SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), "LastChanged": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "BelongsToViews": SimTypeInt(signed=False, label="UInt32"), "NumberOfViews": SimTypeInt(signed=False, label="UInt32"), "ViewInfo": SimTypePointer(SimStruct({"ViewId": SimTypeInt(signed=True, label="Int32"), "NumRoutes": SimTypeInt(signed=False, label="UInt32"), "Route": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestFlags": SimTypeInt(signed=False, label="UInt32"), "HoldRoute": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="_Anonymous_e__Struct", pack=False, align=None), offset=0)}, name="RTM_DEST_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "DestAddress", "ProtocolId", "TargetViews", "DestInfo"]),
        #
        'RtmGetMostSpecificDestination': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"DestHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestAddress": SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), "LastChanged": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "BelongsToViews": SimTypeInt(signed=False, label="UInt32"), "NumberOfViews": SimTypeInt(signed=False, label="UInt32"), "ViewInfo": SimTypePointer(SimStruct({"ViewId": SimTypeInt(signed=True, label="Int32"), "NumRoutes": SimTypeInt(signed=False, label="UInt32"), "Route": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestFlags": SimTypeInt(signed=False, label="UInt32"), "HoldRoute": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="_Anonymous_e__Struct", pack=False, align=None), offset=0)}, name="RTM_DEST_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "DestAddress", "ProtocolId", "TargetViews", "DestInfo"]),
        #
        'RtmGetLessSpecificDestination': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"DestHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestAddress": SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), "LastChanged": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "BelongsToViews": SimTypeInt(signed=False, label="UInt32"), "NumberOfViews": SimTypeInt(signed=False, label="UInt32"), "ViewInfo": SimTypePointer(SimStruct({"ViewId": SimTypeInt(signed=True, label="Int32"), "NumRoutes": SimTypeInt(signed=False, label="UInt32"), "Route": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestFlags": SimTypeInt(signed=False, label="UInt32"), "HoldRoute": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="_Anonymous_e__Struct", pack=False, align=None), offset=0)}, name="RTM_DEST_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "DestHandle", "ProtocolId", "TargetViews", "DestInfo"]),
        #
        'RtmGetExactMatchRoute': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"DestHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "RouteOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Neighbour": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "State": SimTypeChar(label="Byte"), "Flags1": SimTypeChar(label="Byte"), "Flags": SimTypeShort(signed=False, label="UInt16"), "PrefInfo": SimStruct({"Metric": SimTypeInt(signed=False, label="UInt32"), "Preference": SimTypeInt(signed=False, label="UInt32")}, name="RTM_PREF_INFO", pack=False, align=None), "BelongsToViews": SimTypeInt(signed=False, label="UInt32"), "EntitySpecificInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "NextHopsList": SimStruct({"NumNextHops": SimTypeShort(signed=False, label="UInt16"), "NextHops": SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)}, name="RTM_NEXTHOP_LIST", pack=False, align=None)}, name="RTM_ROUTE_INFO", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "DestAddress", "MatchingFlags", "RouteInfo", "InterfaceIndex", "TargetViews", "RouteHandle"]),
        #
        'RtmIsBestRoute': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "RouteHandle", "BestInViews"]),
        #
        'RtmAddNextHop': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"NextHopAddress": SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), "NextHopOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "InterfaceIndex": SimTypeInt(signed=False, label="UInt32"), "State": SimTypeShort(signed=False, label="UInt16"), "Flags": SimTypeShort(signed=False, label="UInt16"), "EntitySpecificInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "RemoteNextHop": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="RTM_NEXTHOP_INFO", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NextHopInfo", "NextHopHandle", "ChangeFlags"]),
        #
        'RtmFindNextHop': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"NextHopAddress": SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), "NextHopOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "InterfaceIndex": SimTypeInt(signed=False, label="UInt32"), "State": SimTypeShort(signed=False, label="UInt16"), "Flags": SimTypeShort(signed=False, label="UInt16"), "EntitySpecificInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "RemoteNextHop": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="RTM_NEXTHOP_INFO", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimTypePointer(SimStruct({"NextHopAddress": SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), "NextHopOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "InterfaceIndex": SimTypeInt(signed=False, label="UInt32"), "State": SimTypeShort(signed=False, label="UInt16"), "Flags": SimTypeShort(signed=False, label="UInt16"), "EntitySpecificInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "RemoteNextHop": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="RTM_NEXTHOP_INFO", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NextHopInfo", "NextHopHandle", "NextHopPointer"]),
        #
        'RtmDeleteNextHop': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"NextHopAddress": SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), "NextHopOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "InterfaceIndex": SimTypeInt(signed=False, label="UInt32"), "State": SimTypeShort(signed=False, label="UInt16"), "Flags": SimTypeShort(signed=False, label="UInt16"), "EntitySpecificInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "RemoteNextHop": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="RTM_NEXTHOP_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NextHopHandle", "NextHopInfo"]),
        #
        'RtmGetNextHopPointer': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimStruct({"NextHopAddress": SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), "NextHopOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "InterfaceIndex": SimTypeInt(signed=False, label="UInt32"), "State": SimTypeShort(signed=False, label="UInt16"), "Flags": SimTypeShort(signed=False, label="UInt16"), "EntitySpecificInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "RemoteNextHop": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="RTM_NEXTHOP_INFO", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NextHopHandle", "NextHopPointer"]),
        #
        'RtmLockNextHop': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypePointer(SimStruct({"NextHopAddress": SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), "NextHopOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "InterfaceIndex": SimTypeInt(signed=False, label="UInt32"), "State": SimTypeShort(signed=False, label="UInt16"), "Flags": SimTypeShort(signed=False, label="UInt16"), "EntitySpecificInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "RemoteNextHop": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="RTM_NEXTHOP_INFO", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NextHopHandle", "Exclusive", "LockNextHop", "NextHopPointer"]),
        #
        'RtmCreateDestEnum': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "TargetViews", "EnumFlags", "NetAddress", "ProtocolId", "RtmEnumHandle"]),
        #
        'RtmGetEnumDests': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"DestHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestAddress": SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), "LastChanged": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "BelongsToViews": SimTypeInt(signed=False, label="UInt32"), "NumberOfViews": SimTypeInt(signed=False, label="UInt32"), "ViewInfo": SimTypePointer(SimStruct({"ViewId": SimTypeInt(signed=True, label="Int32"), "NumRoutes": SimTypeInt(signed=False, label="UInt32"), "Route": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestFlags": SimTypeInt(signed=False, label="UInt32"), "HoldRoute": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="_Anonymous_e__Struct", pack=False, align=None), offset=0)}, name="RTM_DEST_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "EnumHandle", "NumDests", "DestInfos"]),
        #
        'RtmReleaseDests': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"DestHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestAddress": SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), "LastChanged": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "BelongsToViews": SimTypeInt(signed=False, label="UInt32"), "NumberOfViews": SimTypeInt(signed=False, label="UInt32"), "ViewInfo": SimTypePointer(SimStruct({"ViewId": SimTypeInt(signed=True, label="Int32"), "NumRoutes": SimTypeInt(signed=False, label="UInt32"), "Route": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestFlags": SimTypeInt(signed=False, label="UInt32"), "HoldRoute": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="_Anonymous_e__Struct", pack=False, align=None), offset=0)}, name="RTM_DEST_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NumDests", "DestInfos"]),
        #
        'RtmCreateRouteEnum': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"DestHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "RouteOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Neighbour": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "State": SimTypeChar(label="Byte"), "Flags1": SimTypeChar(label="Byte"), "Flags": SimTypeShort(signed=False, label="UInt16"), "PrefInfo": SimStruct({"Metric": SimTypeInt(signed=False, label="UInt32"), "Preference": SimTypeInt(signed=False, label="UInt32")}, name="RTM_PREF_INFO", pack=False, align=None), "BelongsToViews": SimTypeInt(signed=False, label="UInt32"), "EntitySpecificInfo": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "NextHopsList": SimStruct({"NumNextHops": SimTypeShort(signed=False, label="UInt16"), "NextHops": SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)}, name="RTM_NEXTHOP_LIST", pack=False, align=None)}, name="RTM_ROUTE_INFO", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "DestHandle", "TargetViews", "EnumFlags", "StartDest", "MatchingFlags", "CriteriaRoute", "CriteriaInterface", "RtmEnumHandle"]),
        #
        'RtmGetEnumRoutes': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "EnumHandle", "NumRoutes", "RouteHandles"]),
        #
        'RtmReleaseRoutes': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NumRoutes", "RouteHandles"]),
        #
        'RtmCreateNextHopEnum': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "EnumFlags", "NetAddress", "RtmEnumHandle"]),
        #
        'RtmGetEnumNextHops': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "EnumHandle", "NumNextHops", "NextHopHandles"]),
        #
        'RtmReleaseNextHops': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NumNextHops", "NextHopHandles"]),
        #
        'RtmDeleteEnumHandle': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "EnumHandle"]),
        #
        'RtmRegisterForChangeNotification': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "TargetViews", "NotifyFlags", "NotifyContext", "NotifyHandle"]),
        #
        'RtmGetChangedDests': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"DestHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestAddress": SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), "LastChanged": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "BelongsToViews": SimTypeInt(signed=False, label="UInt32"), "NumberOfViews": SimTypeInt(signed=False, label="UInt32"), "ViewInfo": SimTypePointer(SimStruct({"ViewId": SimTypeInt(signed=True, label="Int32"), "NumRoutes": SimTypeInt(signed=False, label="UInt32"), "Route": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestFlags": SimTypeInt(signed=False, label="UInt32"), "HoldRoute": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="_Anonymous_e__Struct", pack=False, align=None), offset=0)}, name="RTM_DEST_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NotifyHandle", "NumDests", "ChangedDests"]),
        #
        'RtmReleaseChangedDests': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"DestHandle": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestAddress": SimStruct({"AddressFamily": SimTypeShort(signed=False, label="UInt16"), "NumBits": SimTypeShort(signed=False, label="UInt16"), "AddrBits": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16)}, name="RTM_NET_ADDRESS", pack=False, align=None), "LastChanged": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "BelongsToViews": SimTypeInt(signed=False, label="UInt32"), "NumberOfViews": SimTypeInt(signed=False, label="UInt32"), "ViewInfo": SimTypePointer(SimStruct({"ViewId": SimTypeInt(signed=True, label="Int32"), "NumRoutes": SimTypeInt(signed=False, label="UInt32"), "Route": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "DestFlags": SimTypeInt(signed=False, label="UInt32"), "HoldRoute": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="_Anonymous_e__Struct", pack=False, align=None), offset=0)}, name="RTM_DEST_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NotifyHandle", "NumDests", "ChangedDests"]),
        #
        'RtmIgnoreChangedDests': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NotifyHandle", "NumDests", "ChangedDests"]),
        #
        'RtmGetChangeStatus': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NotifyHandle", "DestHandle", "ChangeStatus"]),
        #
        'RtmMarkDestForChangeNotification': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NotifyHandle", "DestHandle", "MarkDest"]),
        #
        'RtmIsMarkedForChangeNotification': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NotifyHandle", "DestHandle", "DestMarked"]),
        #
        'RtmDeregisterFromChangeNotification': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NotifyHandle"]),
        #
        'RtmCreateRouteList': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "RouteListHandle"]),
        #
        'RtmInsertInRouteList': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "RouteListHandle", "NumRoutes", "RouteHandles"]),
        #
        'RtmCreateRouteListEnum': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "RouteListHandle", "RtmEnumHandle"]),
        #
        'RtmGetListEnumRoutes': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), label="LPArray", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "EnumHandle", "NumRoutes", "RouteHandles"]),
        #
        'RtmDeleteRouteList': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "RouteListHandle"]),
        #
        'RtmReferenceHandles': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["RtmRegHandle", "NumHandles", "RtmHandles"]),
    }

lib.set_prototypes(prototypes)



from . import MemoryMixin


class InspectMixinHigh(MemoryMixin):
    def store(self, addr, data, size=None, condition=None, endness=None, inspect=True, **kwargs):
        if not inspect or not self.state.supports_inspect:
            super().store(addr, data, size=size, condition=condition, endness=endness, inspect=inspect, **kwargs)
            return

        if self.category == "reg":
            self.state._inspect(
                "reg_write",
                BP_BEFORE,
                reg_write_offset=addr,
                reg_write_length=size,
                reg_write_expr=data,
                reg_write_condition=condition,
                reg_write_endness=endness,
            )
            addr = self.state._inspect_getattr("reg_write_offset", addr)
            size = self.state._inspect_getattr("reg_write_length", size)
            data = self.state._inspect_getattr("reg_write_expr", data)
            condition = self.state._inspect_getattr("reg_write_condition", condition)
            endness = self.state._inspect_getattr("reg_write_endness", endness)
        elif self.category == "mem":
            self.state._inspect(
                "mem_write",
                BP_BEFORE,
                mem_write_address=addr,
                mem_write_length=size,
                mem_write_expr=data,
                mem_write_condition=condition,
                mem_write_endness=endness,
            )
            addr = self.state._inspect_getattr("mem_write_address", addr)
            size = self.state._inspect_getattr("mem_write_length", size)
            data = self.state._inspect_getattr("mem_write_expr", data)
            condition = self.state._inspect_getattr("mem_write_condition", condition)
            endness = self.state._inspect_getattr("mem_write_endness", endness)

        super().store(addr, data, size=size, condition=condition, endness=endness, inspect=inspect, **kwargs)

        if self.category == "reg":
            self.state._inspect(
                "reg_write",
                BP_AFTER,
                reg_write_offset=addr,
                reg_write_length=size,
                reg_write_expr=data,
                reg_write_condition=condition,
                reg_write_endness=endness,
            )
        elif self.category == "mem":
            self.state._inspect(
                "mem_write",
                BP_AFTER,
                mem_write_address=addr,
                mem_write_length=size,
                mem_write_expr=data,
                mem_write_condition=condition,
                mem_write_endness=endness,
            )

    def load(self, addr, size=None, condition=None, endness=None, inspect=True, **kwargs):
        if not inspect or not self.state.supports_inspect:
            return super().load(addr, size=size, condition=condition, endness=endness, inspect=inspect, **kwargs)

        if self.category == "reg":
            self.state._inspect(
                "reg_read",
                BP_BEFORE,
                reg_read_offset=addr,
                reg_read_length=size,
                reg_read_condition=condition,
                reg_read_endness=endness,
            )
            addr = self.state._inspect_getattr("reg_read_offset", addr)
            size = self.state._inspect_getattr("reg_read_length", size)
            condition = self.state._inspect_getattr("reg_read_condition", condition)
            endness = self.state._inspect_getattr("reg_read_endness", endness)
        elif self.category == "mem":
            self.state._inspect(
                "mem_read",
                BP_BEFORE,
                mem_read_address=addr,
                mem_read_length=size,
                mem_read_condition=condition,
                mem_read_endness=endness,
            )
            addr = self.state._inspect_getattr("mem_read_address", addr)
            size = self.state._inspect_getattr("mem_read_length", size)
            condition = self.state._inspect_getattr("mem_read_condition", condition)
            endness = self.state._inspect_getattr("mem_read_endness", endness)

        r = super().load(addr, size=size, condition=condition, endness=endness, inspect=inspect, **kwargs)

        if self.category == "mem":
            self.state._inspect(
                "mem_read",
                BP_AFTER,
                mem_read_expr=r,
                mem_read_address=addr,
                mem_read_length=size,
                mem_read_condition=condition,
                mem_read_endness=endness,
            )
            r = self.state._inspect_getattr("mem_read_expr", r)

        elif self.category == "reg":
            self.state._inspect(
                "reg_read",
                BP_AFTER,
                reg_read_expr=r,
                reg_read_offset=addr,
                reg_read_length=size,
                reg_read_condition=condition,
                reg_read_endness=endness,
            )
            r = self.state._inspect_getattr("reg_read_expr", r)

        return r

    def _add_constraints(self, c, add_constraints=True, inspect=True, **kwargs):
        if inspect and self.state.supports_inspect:
            # tracer uses address_concretization_add_constraints
            add_constraints = self.state._inspect_getattr("address_concretization_add_constraints", add_constraints)

        super()._add_constraints(c, add_constraints=add_constraints, inspect=inspect, **kwargs)


from ...state_plugins.inspect import BP_BEFORE, BP_AFTER



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("dssec.dll")
prototypes = \
    {
        #
        'DSCreateISecurityInfoObject': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ISecurityInformation"), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimStruct({"Revision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "Control": SimTypeShort(signed=False, label="UInt16"), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Group": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Sacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0), "Dacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0)}, name="SECURITY_DESCRIPTOR", pack=False, align=None), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Revision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "Control": SimTypeShort(signed=False, label="UInt16"), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Group": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Sacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0), "Dacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0)}, name="SECURITY_DESCRIPTOR", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pwszObjectPath", "pwszObjectClass", "dwFlags", "ppSI", "pfnReadSD", "pfnWriteSD", "lpContext"]),
        #
        'DSCreateISecurityInfoObjectEx': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ISecurityInformation"), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimStruct({"Revision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "Control": SimTypeShort(signed=False, label="UInt16"), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Group": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Sacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0), "Dacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0)}, name="SECURITY_DESCRIPTOR", pack=False, align=None), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Revision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "Control": SimTypeShort(signed=False, label="UInt16"), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Group": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Sacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0), "Dacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0)}, name="SECURITY_DESCRIPTOR", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pwszObjectPath", "pwszObjectClass", "pwszServer", "pwszUserName", "pwszPassword", "dwFlags", "ppSI", "pfnReadSD", "pfnWriteSD", "lpContext"]),
        #
        'DSCreateSecurityPage': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimStruct({"Revision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "Control": SimTypeShort(signed=False, label="UInt16"), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Group": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Sacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0), "Dacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0)}, name="SECURITY_DESCRIPTOR", pack=False, align=None), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Revision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "Control": SimTypeShort(signed=False, label="UInt16"), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Group": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Sacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0), "Dacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0)}, name="SECURITY_DESCRIPTOR", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pwszObjectPath", "pwszObjectClass", "dwFlags", "phPage", "pfnReadSD", "pfnWriteSD", "lpContext"]),
        #
        'DSEditSecurity': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimStruct({"Revision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "Control": SimTypeShort(signed=False, label="UInt16"), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Group": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Sacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0), "Dacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0)}, name="SECURITY_DESCRIPTOR", pack=False, align=None), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Revision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "Control": SimTypeShort(signed=False, label="UInt16"), "Owner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Group": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "Sacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0), "Dacl": SimTypePointer(SimStruct({"AclRevision": SimTypeChar(label="Byte"), "Sbz1": SimTypeChar(label="Byte"), "AclSize": SimTypeShort(signed=False, label="UInt16"), "AceCount": SimTypeShort(signed=False, label="UInt16"), "Sbz2": SimTypeShort(signed=False, label="UInt16")}, name="ACL", pack=False, align=None), offset=0)}, name="SECURITY_DESCRIPTOR", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hwndOwner", "pwszObjectPath", "pwszObjectClass", "dwFlags", "pwszCaption", "pfnReadSD", "pfnWriteSD", "lpContext"]),
    }

lib.set_prototypes(prototypes)



import angr
import logging

l = logging.getLogger(name=__name__)


class strncat(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, dst, src, limit):
        strncpy = angr.SIM_PROCEDURES["libc"]["strncpy"]
        strlen = angr.SIM_PROCEDURES["libc"]["strlen"]
        dst_len = self.inline_call(strlen, dst).ret_expr
        src_len = self.inline_call(strlen, src).ret_expr

        self.inline_call(strncpy, dst + dst_len, src, limit, src_len=src_len)

        return dst



# pylint: disable=missing-class-docstring
from angr import SimProcedure


class ExFreePoolWithTag(SimProcedure):
    def run(self, P, Tag):  # pylint:disable=arguments-differ, unused-argument
        self.state.heap._free(P)
